//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Frame/Fit2DFrame.cpp
//! @brief     Implements class Fit2DFrame.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Frame/Fit2DFrame.h"
#include "Base/Util/Assert.h"
#include "Device/Data/Datafield.h"
#include "Device/IO/DiffUtil.h"
#include "GUI/Model/Data/Data1DItem.h"
#include "GUI/Model/Data/Data2DItem.h"
#include "GUI/Model/Project/ProjectDocument.h"
#include "GUI/Support/Util/ActionFactory.h"
#include "GUI/View/Plot2D/ColorMapCanvas.h"
#include "GUI/View/PlotComparison/FitFlowWidget.h"
#include "GUI/View/PlotScale/RangeUtil.h"
#include "GUI/View/PlotUtil/ColorMap.h"
#include "GUI/View/PlotUtil/PlotStatusLabel.h"
#include "GUI/View/Setup/Scale2DEditor.h"
#include <QAction>
#include <QVBoxLayout>

Fit2DFrame::Fit2DFrame(QWidget* parent)
    : DataAccessWidget(parent)
    , m_realCanvas(new ColorMapCanvas(this))
    , m_simuCanvas(new ColorMapCanvas(this))
    , m_diffCanvas(new ColorMapCanvas(this))
    , m_fitFlowWidget(new FitFlowWidget(this))
    , m_statusLabel(new PlotStatusLabel(nullptr, this))
    , m_propertyWidget(new Scale2DEditor(this))
    , m_resetViewAction(new QAction(this))
{
    auto* vlayout = new QVBoxLayout;
    vlayout->setSpacing(0);

    m_propertyWidget->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Minimum);

    auto* gridLayout = new QGridLayout;
    gridLayout->setSpacing(0);

    gridLayout->addWidget(m_realCanvas, 0, 0);
    gridLayout->addWidget(m_simuCanvas, 0, 1);
    gridLayout->addWidget(m_diffCanvas, 1, 0);
    gridLayout->addWidget(m_fitFlowWidget, 1, 1);

    vlayout->addLayout(gridLayout);
    vlayout->addWidget(m_statusLabel);

    m_statusLabel->reset();
    m_statusLabel->addPlot(m_realCanvas->colorMap());
    m_statusLabel->addPlot(m_simuCanvas->colorMap());
    m_statusLabel->addPlot(m_diffCanvas->colorMap());

    auto* hlayout = new QHBoxLayout;
    hlayout->setSpacing(0);
    hlayout->addLayout(vlayout);
    hlayout->addWidget(m_propertyWidget);
    setLayout(hlayout);

    m_resetViewAction->setText("Center view");
    m_resetViewAction->setIcon(QIcon(":/images/camera-metering-center.svg"));
    m_resetViewAction->setToolTip("Reset View");
    connect(m_resetViewAction, &QAction::triggered, this, &Fit2DFrame::onResetViewAction,
            Qt::UniqueConnection);

    m_togglePropertiesAction =
        ActionFactory::createTogglePropertiesPanelAction(this, m_propertyWidget);

    m_propertyWidget->setVisible(false);
    connect(m_propertyWidget, &DataPropertyWidget::axesRangeResetRequested, this,
            &Fit2DFrame::onResetViewAction, Qt::UniqueConnection);
}

void Fit2DFrame::setJobOrDatafileItem(QObject* job_item)
{
    JobItem* oldJob = jobItem();

    setBaseItem(job_item);
    ASSERT(jobItem());

    if (oldJob != jobItem())
        GUI::View::RangeUtil::setCommonRangeZ(mainData2DItems());
    updateDiffData();
    connectItems();

    m_simuCanvas->itemToCanvas(simuData2DItem());
    m_realCanvas->itemToCanvas(realData2DItem());
    m_diffCanvas->itemToCanvas(diffData2DItem());

    m_fitFlowWidget->setJobOrDatafileItem(job_item);
    m_propertyWidget->setJobOrDatafileItem(job_item);
}

QList<QAction*> Fit2DFrame::actionList()
{
    return {m_resetViewAction, m_togglePropertiesAction};
}

void Fit2DFrame::onResetViewAction()
{
    ASSERT(simuData2DItem() && diffData2DItem() && realData2DItem());
    simuData2DItem()->resetView();
    realData2DItem()->resetView();
    diffData2DItem()->resetView();

    // synchronize data range between simulated and real
    GUI::View::RangeUtil::setCommonRangeZ(mainData2DItems());
    gProjectDocument.value()->setModified();
}

void Fit2DFrame::connectItems()
{
    // sync XY view area between simulated, real and difference plots
    for (auto senderItem : allData2DItems())
        for (auto receiverItem : allData2DItems())
            if (receiverItem != senderItem)
                connect(senderItem, &DataItem::updateOtherPlots, receiverItem,
                        &DataItem::checkXYranges, Qt::UniqueConnection);

    // sync Z range between simulated and real
    connect(simuData2DItem(), &Data2DItem::alignRanges,
            [this] { GUI::View::RangeUtil::setCommonRangeZ(mainData2DItems()); });

    // sync Z range: simu --> real
    connect(simuData2DItem(), &DataItem::updateOtherPlots, realData2DItem(),
            &Data2DItem::copyZRangeFromItem, Qt::UniqueConnection);

    // sync Z range: real --> simu
    connect(realData2DItem(), &DataItem::updateOtherPlots, simuData2DItem(),
            &Data2DItem::copyZRangeFromItem, Qt::UniqueConnection);

    // update diff data if simulation data changes
    connect(simuData2DItem(), &Data1DItem::datafieldChanged, this, &Fit2DFrame::updateDiffData,
            Qt::UniqueConnection);
}

void Fit2DFrame::updateDiffData()
{
    ASSERT(simuData2DItem() && diffData2DItem() && realData2DItem());
    if (!simuData2DItem()->c_field() || !realData2DItem()->c_field())
        return;

    diffData2DItem()->setDatafield(DiffUtil::relativeDifferenceField(*simuData2DItem()->c_field(),
                                                                     *realData2DItem()->c_field()));

    // keep Z axis range up with data range
    diffData2DItem()->computeDataRange();
}
