//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Frame/MaskEditingFrame.cpp
//! @brief     Implements class MaskEditingFrame.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Frame/MaskEditingFrame.h"
#include "GUI/Model/Data/Data2DItem.h"
#include "GUI/Model/Mask/MaskContainerModel.h"
#include "GUI/Model/Mask/MaskItems.h"
#include "GUI/View/Mask/MaskEditorActions.h"
#include "GUI/View/Mask/MaskEditorCanvas.h"
#include "GUI/View/Mask/MaskEditorToolbar.h"
#include "GUI/View/Mask/MaskGraphicsScene.h"
#include "GUI/View/Setup/MaskEditorPropertyPanel.h"
#include <QSplitter>

MaskEditingFrame::MaskEditingFrame(QWidget* parent)
    : DataAccessWidget(parent)
    , m_editorActions(new MaskEditorActions(this))
    , m_editorPropertyPanel(new MaskEditorPropertyPanel)
    , m_editorCanvas(new MaskEditorCanvas)
{
    setObjectName("MaskEditor");
    setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);

    auto* splitter = new QSplitter;
    splitter->addWidget(m_editorCanvas);
    splitter->addWidget(m_editorPropertyPanel);
    splitter->setCollapsible(1, true);

    QHBoxLayout* mainLayout = new QHBoxLayout(this);
    mainLayout->addWidget(splitter);
    mainLayout->setSpacing(0);

    auto* toolbar = new MaskEditorToolbar(m_editorActions);
    toolbar->setOrientation(Qt::Vertical);
    toolbar->setStyleSheet("QToolBar{border-left:1px solid rgb(180,180,180);}");
    mainLayout->addWidget(toolbar);

    m_editorPropertyPanel->setPanelHidden(true);

    //... Connections

    // reset view request is propagated from editorActions to graphics view
    connect(m_editorActions, &MaskEditorActions::resetViewRequest, m_editorCanvas,
            &MaskEditorCanvas::onResetViewRequest);

    // tool panel request is propagated from editorActions to this MaskEditor
    connect(m_editorActions, &MaskEditorActions::propertyPanelRequest, this,
            &MaskEditingFrame::onPropertyPanelRequest);

    // save plot request is propagated from editorActions to graphics scene
    connect(m_editorActions, &MaskEditorActions::savePlotRequest, m_editorCanvas,
            &MaskEditorCanvas::onSavePlotRequest);

    // selection/drawing activity is propagated from Toolbar to graphics scene
    connect(toolbar, &MaskEditorToolbar::activityChanged, m_editorCanvas->getScene(),
            &MaskGraphicsScene::onActivityChanged);

    // mask value is propagated from Toolbar to graphics scene
    connect(toolbar, &MaskEditorToolbar::maskValueChanged, m_editorCanvas->getScene(),
            &MaskGraphicsScene::onMaskValueChanged);

    // show results request is propagated from Toolbar to Canvas
    connect(toolbar, &MaskEditorToolbar::presentationChange, m_editorCanvas,
            &MaskEditorCanvas::onPresentationChange);

    // space bar push (request for zoom mode) is propagated from graphics view to Toolbar
    connect(m_editorCanvas, &MaskEditorCanvas::changeActivityRequest, toolbar,
            &MaskEditorToolbar::onChangeActivityRequest);

    // Delete request is propagated from canvas to actions
    connect(m_editorCanvas, &MaskEditorCanvas::deleteSelectedRequest, m_editorActions,
            &MaskEditorActions::onDeleteMaskAction);

    // context menu request is propagated from graphics scene to MaskEditorActions
    connect(m_editorCanvas->getScene(), &MaskGraphicsScene::itemContextMenuRequest, m_editorActions,
            &MaskEditorActions::onItemContextMenuRequest);

    // context menu request is propagated from PropertyPanel to MaskEditorActions
    connect(m_editorPropertyPanel, &MaskEditorPropertyPanel::itemContextMenuRequest,
            m_editorActions, &MaskEditorActions::onItemContextMenuRequest);
}

void MaskEditingFrame::setJobOrDatafileItem(QObject* job_or_real_item)
{
    setBaseItem(job_or_real_item);
    m_editorPropertyPanel->setJobOrDatafileItem(job_or_real_item);
    setContextMW();
}

//! shows/hides right panel with properties
void MaskEditingFrame::onPropertyPanelRequest()
{
    m_editorPropertyPanel->setPanelHidden(!m_editorPropertyPanel->isHidden());
}

//! Returns list of actions intended for styled toolbar (on the top).

QList<QAction*> MaskEditingFrame::actionList()
{
    return m_editorActions->topToolbarActions();
}

void MaskEditingFrame::showEvent(QShowEvent*)
{
    if (currentData2DItem())
        setContextMW();
}

void MaskEditingFrame::hideEvent(QHideEvent*)
{
    m_editorPropertyPanel->resetContextMPP();
    m_editorCanvas->resetContextMEC();
}

void MaskEditingFrame::setContextMW()
{
    ASSERT(currentData2DItem());

    MaskContainerModel* containerModel = currentData2DItem()->getOrCreateMaskContainerModel();
    ASSERT(containerModel);

    m_editorPropertyPanel->setContextMPP(containerModel);

    m_editorCanvas->setContextMEC(currentData2DItem(), m_editorPropertyPanel->selectionModel());

    m_editorActions->setModels(containerModel, m_editorPropertyPanel->selectionModel());

    update();
}
