//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Instrument/InstrumentListing.cpp
//! @brief     Implements class InstrumentListing.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Instrument/InstrumentListing.h"
#include "GUI/Model/Project/ProjectDocument.h"
#include "GUI/Support/Tool/Globals.h"
#include "GUI/View/Instrument/InstrumentLibraryEditor.h"
#include "GUI/View/Instrument/InstrumentListModel.h"
#include <QAction>
#include <QMessageBox>
#include <QVBoxLayout>

InstrumentListing::InstrumentListing(ProjectDocument* document)
    : m_document(document)
{
    m_instrumentLibrary.load();

    setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Expanding);
    setViewMode(QListView::IconMode);
    setIconSize(QSize(96, 84));
    setMovement(QListView::Static);
    setMaximumWidth(200);
    setSelectionMode(QAbstractItemView::SingleSelection);

    m_model = new InstrumentListModel(this, m_document->multiNotifier());
    setModel(m_model);

    m_newGisasAction = new QAction("New GISAS", this);
    m_newGisasAction->setIcon(QIcon(":/images/shape-square-plus.svg"));
    m_newGisasAction->setToolTip("Add new GISAS instrument with default settings");
    connect(m_newGisasAction, &QAction::triggered, this, &InstrumentListing::onNewGisas);
    addAction(m_newGisasAction);

    m_newOffspecAction = new QAction("New off-specular", this);
    m_newOffspecAction->setIcon(QIcon(":/images/shape-square-plus.svg"));
    m_newOffspecAction->setToolTip("Add new off-specular instrument with default settings");
    connect(m_newOffspecAction, &QAction::triggered, this, &InstrumentListing::onNewOffspec);
    addAction(m_newOffspecAction);

    m_newSpecularAction = new QAction("New specular", this);
    m_newSpecularAction->setIcon(QIcon(":/images/shape-square-plus.svg"));
    m_newSpecularAction->setToolTip("Add new specular instrument with default settings");
    connect(m_newSpecularAction, &QAction::triggered, this, &InstrumentListing::onNewSpecular);
    addAction(m_newSpecularAction);

    m_newDepthprobeAction = new QAction("New depth probe", this);
    m_newDepthprobeAction->setIcon(QIcon(":/images/shape-square-plus.svg"));
    m_newDepthprobeAction->setToolTip("Add new depth probe instrument with default settings");
    connect(m_newDepthprobeAction, &QAction::triggered, this, &InstrumentListing::onNewDepthprobe);
    addAction(m_newDepthprobeAction);

    m_separatorAction1 = new QAction(this);
    m_separatorAction1->setSeparator(true);
    addAction(m_separatorAction1);

    m_removeAction = new QAction("Remove", this);
    m_removeAction->setIcon(QIcon(":/images/delete.svg"));
    m_removeAction->setToolTip("Remove selected instrument");
    connect(m_removeAction, &QAction::triggered, this, &InstrumentListing::onRemove);
    addAction(m_removeAction);

    m_copyAction = new QAction("Copy", this);
    m_copyAction->setIcon(QIcon(":/images/content-copy.svg"));
    m_copyAction->setToolTip("Make a copy of the selected instrument");
    connect(m_copyAction, &QAction::triggered, this, &InstrumentListing::onCopy);
    addAction(m_copyAction);

    m_separatorAction2 = new QAction(this);
    m_separatorAction2->setSeparator(true);
    addAction(m_separatorAction2);

    m_storeInLibraryAction = new QAction("Store in library", this);
    m_storeInLibraryAction->setIcon(QIcon(":/images/library.svg"));
    m_storeInLibraryAction->setToolTip("Store instrument in library");
    connect(m_storeInLibraryAction, &QAction::triggered, this,
            &InstrumentListing::onStoreInLibrary);
    addAction(m_storeInLibraryAction);

    m_loadFromLibraryAction = new QAction("Choose from library", this);
    m_loadFromLibraryAction->setIcon(QIcon(":/images/library.svg"));
    m_loadFromLibraryAction->setToolTip("Load an instrument from the instrument library");
    connect(m_loadFromLibraryAction, &QAction::triggered, this,
            &InstrumentListing::onLoadFromLibrary);
    addAction(m_loadFromLibraryAction);

    setContextMenuPolicy(Qt::ActionsContextMenu);

    updateActions();
    restoreSelection();

    connect(selectionModel(), &QItemSelectionModel::selectionChanged, this,
            &InstrumentListing::onItemSelectionChanged);
}

InstrumentListing::~InstrumentListing()
{
    m_instrumentLibrary.saveIfModified();
}

QSize InstrumentListing::sizeHint() const
{
    return QSize(170, 400);
}

QSize InstrumentListing::minimumSizeHint() const
{
    return QSize(96, 200);
}

QList<QAction*> InstrumentListing::toolbarActions() const
{
    return {m_newGisasAction,       m_newOffspecAction, m_newSpecularAction,
            m_newDepthprobeAction,  m_separatorAction1, m_removeAction,
            m_copyAction,           m_separatorAction2, m_storeInLibraryAction,
            m_loadFromLibraryAction};
}

InstrumentItem* InstrumentListing::currentInstrumentItem() const
{
    const QModelIndexList indexes = selectionModel()->selectedIndexes();
    if (!indexes.empty())
        return m_model->instrumentItemForIndex(indexes.front());
    return nullptr;
}

void InstrumentListing::onItemSelectionChanged()
{
    updateActions();

    QModelIndexList indexes = selectionModel()->selectedIndexes();
    if (!indexes.empty()) {
        QModelIndex current = indexes.front();
        m_document->instrumentModel()->setSelectedIndex(current.row());
        emit instrumentSelected(m_model->instrumentItemForIndex(current));
    } else
        emit instrumentSelected(nullptr);
}

void InstrumentListing::onNewGisas()
{
    QModelIndex idx = m_model->addNewGISASInstrument();
    selectionModel()->select(idx, QItemSelectionModel::ClearAndSelect);
}

void InstrumentListing::onNewOffspec()
{
    QModelIndex idx = m_model->addNewOffspecInstrument();
    selectionModel()->select(idx, QItemSelectionModel::ClearAndSelect);
}

void InstrumentListing::onNewSpecular()
{
    QModelIndex idx = m_model->addNewSpecularInstrument();
    selectionModel()->select(idx, QItemSelectionModel::ClearAndSelect);
}

void InstrumentListing::onNewDepthprobe()
{
    QModelIndex idx = m_model->addNewDepthprobeInstrument();
    selectionModel()->select(idx, QItemSelectionModel::ClearAndSelect);
}

//! Removes currently selected instrument.
void InstrumentListing::onRemove()
{
    QModelIndexList indexes = selectionModel()->selectedIndexes();
    if (!indexes.empty()) {
        m_model->removeInstrument(indexes.front());

        ensureItemSelected();
    }
}

//! Makes a copy of the currently selected instrument.
void InstrumentListing::onCopy()
{
    QModelIndexList indexes = selectionModel()->selectedIndexes();
    if (!indexes.empty()) {
        QModelIndex idx = m_model->copyInstrument(indexes.front());
        selectionModel()->select(idx, QItemSelectionModel::ClearAndSelect);
    }
}

void InstrumentListing::onStoreInLibrary()
{
    if (!selectionModel()->hasSelection())
        return;

    QModelIndex idx = selectionModel()->selectedIndexes().front();
    InstrumentItem* instrument = m_model->instrumentItemForIndex(idx);

    InstrumentLibraryEditor dlg(GUI::Global::mainWindow, &m_instrumentLibrary);
    dlg.execAdd(*instrument);
}

void InstrumentListing::onLoadFromLibrary()
{
    if (m_instrumentLibrary.isEmpty()) {
        QMessageBox::information(GUI::Global::mainWindow, "Select from library",
                                 "The library does not contain instruments so far.");
        return;
    }

    InstrumentLibraryEditor dlg(GUI::Global::mainWindow, &m_instrumentLibrary);

    auto* instrumentToCopy = dlg.execChoose();
    if (instrumentToCopy == nullptr)
        return;

    QModelIndex idx = m_model->copyInstrument(instrumentToCopy);
    selectionModel()->select(idx, QItemSelectionModel::ClearAndSelect);
}

void InstrumentListing::updateActions()
{
    bool enabled = selectionModel()->hasSelection();
    m_removeAction->setEnabled(enabled);
    m_copyAction->setEnabled(enabled);
    m_storeInLibraryAction->setEnabled(enabled);
}

void InstrumentListing::ensureItemSelected()
{
    if (!selectionModel()->hasSelection() && m_model->rowCount()) {
        QModelIndex last = m_model->index(m_model->rowCount() - 1, 0, QModelIndex());
        selectionModel()->select(last, QItemSelectionModel::ClearAndSelect);
    }
}

void InstrumentListing::restoreSelection()
{
    int lastUsed = m_document->instrumentModel()->selectedIndex();
    if (lastUsed >= 0 && lastUsed < m_model->rowCount()) {
        QModelIndex lastUsedIndex = m_model->index(lastUsed, 0, QModelIndex());
        selectionModel()->select(lastUsedIndex, QItemSelectionModel::ClearAndSelect);
    } else
        ensureItemSelected();

    updateActions();
}
