//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Project/SimulationView.cpp
//! @brief     Implements class SimulationView.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Project/SimulationView.h"
#include "GUI/Model/Data/DataItem.h"
#include "GUI/Model/Device/DatafileItem.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Job/JobItem.h"
#include "GUI/Model/Model/DatafilesModel.h"
#include "GUI/Model/Model/JobModel.h"
#include "GUI/Model/Project/ProjectDocument.h"
#include "GUI/Model/Sample/SampleItem.h"
#include "GUI/Model/Sample/SampleValidator.h"
#include "GUI/Support/Data/SimulationOptionsItem.h"
#include "GUI/Support/Tool/Globals.h"
#include "GUI/View/Project/ProjectManager.h"
#include "GUI/View/Project/PythonScriptWidget.h"
#include "GUI/View/Widget/GroupBoxes.h"
#include <QButtonGroup>
#include <QCheckBox>
#include <QFormLayout>
#include <QMessageBox>
#include <QVBoxLayout>
#include <thread>

SimulationView::SimulationView(QWidget* parent, ProjectDocument* document)
    : QWidget(parent)
    , m_document(document)
{
    auto* layout = new QVBoxLayout;
    setLayout(layout);

    //... Data selection

    auto* g1 = new StaticGroupBox("Data selection", this);
    layout->addWidget(g1);
    auto* f1 = new QFormLayout;
    g1->body()->setLayout(f1);
    m_instrumentLabel = new QLabel("Instrument");
    m_instrumentCombo = new QComboBox;
    f1->addRow(m_instrumentLabel, m_instrumentCombo);
    m_sampleCombo = new QComboBox;
    f1->addRow("Sample", m_sampleCombo);
    m_realDataCombo = new QComboBox;
    f1->addRow("Data", m_realDataCombo);

    //... Model evaluation settings

    auto* g2 = new StaticGroupBox("Model evaluation settings", this);
    layout->addWidget(g2);
    auto* l2 = new QVBoxLayout;
    g2->body()->setLayout(l2);

    // analytical vs MC
    auto* b_mc = new QButtonGroup;
    m_analyticalRadio = new QRadioButton("Deterministic computation at pixel center");
    b_mc->addButton(m_analyticalRadio);
    l2->addWidget(m_analyticalRadio);
    m_monteCarloRadio = new QRadioButton("Monte-Carlo computation; number of points per pixel: ");
    b_mc->addButton(m_monteCarloRadio);
    auto* l_mc2 = new QHBoxLayout;
    l2->addLayout(l_mc2);
    l_mc2->addWidget(m_monteCarloRadio);
    m_numberOfMonteCarloPoints = new QSpinBox;
    l_mc2->addWidget(m_numberOfMonteCarloPoints);
    l_mc2->addStretch(1);
    l2->addSpacerItem(new QSpacerItem(10, 10));

    // layer material
    auto* b_lm = new QButtonGroup;
    m_ambientLayerRadio = new QRadioButton("R&T computed for layer material as given");
    b_lm->addButton(m_ambientLayerRadio);
    l2->addWidget(m_ambientLayerRadio);
    m_averageLayerRadio = new QRadioButton("R&T computed for averaged layer material");
    b_lm->addButton(m_averageLayerRadio);
    l2->addWidget(m_averageLayerRadio);
    l2->addSpacerItem(new QSpacerItem(10, 10));

    // also specular peak?
    m_includeSpecularCheck = new QCheckBox("Also compute specular peak");
    l2->addWidget(m_includeSpecularCheck);

    //... Execution settings

    auto* g3 = new StaticGroupBox("Execution settings", this);
    layout->addWidget(g3);
    auto* l3 = new QVBoxLayout;
    g3->body()->setLayout(l3);

    // immediately/background?
    auto* b_exec = new QButtonGroup;
    m_runPolicyImmediatelyRadio = new QRadioButton("Run immediately");
    b_exec->addButton(m_runPolicyImmediatelyRadio);
    l3->addWidget(m_runPolicyImmediatelyRadio);
    m_runPolicyBackgroundRadio = new QRadioButton("Run in background");
    b_exec->addButton(m_runPolicyBackgroundRadio);
    l2->addWidget(m_runPolicyBackgroundRadio);
    l3->addSpacerItem(new QSpacerItem(10, 10));

    // number of threads
    auto* f3 = new QFormLayout;
    m_numberOfThreadsCombo = new QComboBox;
    m_numberOfThreadsCombo->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
    f3->addRow("Number of threads", m_numberOfThreadsCombo);
    l3->addLayout(f3);
    l3->addSpacerItem(new QSpacerItem(10, 10));

    auto* l_launchers = new QHBoxLayout;
    l3->addLayout(l_launchers);
    m_simulateButton = new QPushButton("Run simulation");
    l_launchers->addWidget(m_simulateButton);
    m_exportToPyScriptButton = new QPushButton("Export to Python Script");
    l_launchers->addWidget(m_exportToPyScriptButton);
    l_launchers->addStretch(1);

    layout->addStretch(1);

    //... Set data

    // -- fill combo for "number of threads"
    const int nthreads = static_cast<int>(std::thread::hardware_concurrency());
    m_numberOfThreadsCombo->addItem(QString("Max (%1 threads)").arg(nthreads), nthreads);
    for (int i = nthreads - 1; i > 1; i--)
        m_numberOfThreadsCombo->addItem(QString("%1 threads").arg(i), i);
    m_numberOfThreadsCombo->addItem("1 thread", 1);

    //... Connect

    connect(m_instrumentCombo, &QComboBox::currentTextChanged, [this] { updateStateFromUI(); });
    connect(m_sampleCombo, &QComboBox::currentTextChanged, [this] { updateStateFromUI(); });
    connect(m_realDataCombo, &QComboBox::currentTextChanged, [this] { updateStateFromUI(); });

    connect(m_simulateButton, &QPushButton::clicked, this, &SimulationView::simulate);
    connect(m_exportToPyScriptButton, &QPushButton::clicked, this,
            &SimulationView::exportPythonScript);
    connect(b_exec, &QButtonGroup::buttonClicked, this, &SimulationView::updateEnabling);

    connect(m_runPolicyImmediatelyRadio, &QRadioButton::toggled, [this]() { updateStateFromUI(); });

    connect(m_analyticalRadio, &QRadioButton::toggled, [this]() { updateStateFromUI(); });

    connect(m_averageLayerRadio, &QRadioButton::toggled, [this]() { updateStateFromUI(); });

    connect(m_numberOfThreadsCombo, QOverload<int>::of(&QComboBox::currentIndexChanged),
            [this]() { updateStateFromUI(); });

    connect(m_numberOfMonteCarloPoints, QOverload<int>::of(&QSpinBox::valueChanged),
            [this]() { updateStateFromUI(); });

    connect(m_includeSpecularCheck, &QCheckBox::toggled, [this]() { updateStateFromUI(); });
}

void SimulationView::showEvent(QShowEvent*)
{
    writeOptionsToUI();
}

void SimulationView::writeOptionsToUI()
{
    QSignalBlocker b1(m_runPolicyImmediatelyRadio);
    QSignalBlocker b2(m_analyticalRadio);
    QSignalBlocker b3(m_averageLayerRadio);
    QSignalBlocker b4(m_numberOfThreadsCombo);
    QSignalBlocker b5(m_numberOfMonteCarloPoints);
    QSignalBlocker b6(m_includeSpecularCheck);

    // -- selection group
    updateSelection(m_instrumentCombo, m_document->instrumentModel()->instrumentNames(),
                    optionsItem()->selectedInstrumentIndex());
    updateSelection(m_sampleCombo, m_document->sampleModel()->sampleNames(),
                    optionsItem()->selectedSampleIndex());
    updateSelection(m_realDataCombo, m_document->realModel()->realItemNames(),
                    optionsItem()->selectedDataIndex(), true);

    // -- options group
    optionsItem()->runImmediately() ? m_runPolicyImmediatelyRadio->setChecked(true)
                                    : m_runPolicyBackgroundRadio->setChecked(true);

    optionsItem()->useAnalytical() ? m_analyticalRadio->setChecked(true)
                                   : m_monteCarloRadio->setChecked(true);

    optionsItem()->useAverageMaterials() ? m_averageLayerRadio->setChecked(true)
                                         : m_ambientLayerRadio->setChecked(true);

    m_numberOfThreadsCombo->setCurrentIndex(
        m_numberOfThreadsCombo->findData(optionsItem()->numberOfThreads()));
    m_numberOfMonteCarloPoints->setValue(optionsItem()->numberOfMonteCarloPoints());
    m_includeSpecularCheck->setChecked(optionsItem()->includeSpecularPeak());

    updateEnabling();
}

void SimulationView::hideEvent(QHideEvent*)
{
    // Storing the options has do be done here because on a showEvent the values are set again from
    // the SimulationOptionsItem (=> keep values when switching views)
    readOptionsFromUI();
}

void SimulationView::updateStateFromUI()
{
    readOptionsFromUI();
    m_document->setModified();
}

void SimulationView::simulate()
{
    readOptionsFromUI();
    if (const QString msg = validateSimulationSetup(true); !msg.isEmpty()) {
        QMessageBox::warning(this, "Cannot run the job",
                             "Cannot run the job with current settings\n\n" + msg);
        return;
    }
    JobModel* jobModel = m_document->jobModel();
    JobItem* jobItem = jobModel->addJobItem(selectedSampleItem(), selectedInstrumentItem(),
                                            selectedDatafileItem(), *optionsItem());
    jobModel->runJob(jobItem);
    gProjectDocument.value()->setModified();
}

void SimulationView::exportPythonScript()
{
    readOptionsFromUI();
    if (const QString msg = validateSimulationSetup(false); !msg.isEmpty()) {
        QMessageBox::warning(this, "Cannot export to Python",
                             "Cannot export to Python with current settings\n\n" + msg);
        return;
    }
    auto* pythonWidget = new PythonScriptWidget(GUI::Global::mainWindow);
    pythonWidget->show();
    pythonWidget->raise();
    pythonWidget->generatePythonScript(*selectedSampleItem(), *selectedInstrumentItem(),
                                       *optionsItem(), ProjectManager::instance()->projectDir());
}

void SimulationView::readOptionsFromUI()
{
    optionsItem()->setSelectedInstrumentIndex(m_instrumentCombo->currentIndex());
    optionsItem()->setSelectedSampleIndex(m_sampleCombo->currentIndex());
    optionsItem()->setSelectedDataIndex(m_realDataCombo->currentIndex());

    optionsItem()->setRunImmediately(m_runPolicyImmediatelyRadio->isChecked());
    optionsItem()->setNumberOfThreads(m_numberOfThreadsCombo->currentData().toInt());
    if (m_analyticalRadio->isChecked())
        optionsItem()->setUseAnalytical();
    else
        optionsItem()->setUseMonteCarloIntegration(m_numberOfMonteCarloPoints->value());
    optionsItem()->setUseAverageMaterials(m_averageLayerRadio->isChecked());
    optionsItem()->setIncludeSpecularPeak(m_includeSpecularCheck->isChecked());
}

void SimulationView::updateEnabling()
{
    m_numberOfMonteCarloPoints->setEnabled(m_monteCarloRadio->isChecked());
}

void SimulationView::updateSelection(QComboBox* comboBox, QStringList itemList, int currentIndex,
                                     bool allowNone)
{
    QSignalBlocker b(comboBox);
    const QString previousItem = comboBox->currentText();

    comboBox->clear();
    comboBox->setEnabled(!itemList.isEmpty());

    if (itemList.isEmpty()) {
        comboBox->addItem("Not yet defined");
        return;
    }

    if (allowNone)
        itemList.prepend("None");

    comboBox->addItems(itemList);
    if (currentIndex > 0 && currentIndex < itemList.count())
        comboBox->setCurrentIndex(currentIndex);
    else if (itemList.contains(previousItem))
        comboBox->setCurrentIndex(itemList.indexOf(previousItem));
}

QString SimulationView::validateSimulationSetup(bool validateRealData) const
{
    QString messages;
    const auto append = [&](const QString& m) { messages.append("- " + m + "\n"); };

    if (!selectedSampleItem())
        append("No sample selected");
    else {
        QString m = SampleValidator::isValidSample(selectedSampleItem());
        if (!m.isEmpty())
            append(m);
    }

    if (!selectedInstrumentItem())
        append("No instrument selected");

    if (validateRealData && selectedDatafileItem() && selectedInstrumentItem()
        && !selectedInstrumentItem()->alignedWith(selectedDatafileItem()))
        append("The experimental data does not fit in the selected instrument. Try linking "
               "them in Import Tab.");

    if (selectedDatafileItem()
        && (!selectedDatafileItem()->dataItem() || !selectedDatafileItem()->dataItem()->c_field()))
        append("The experimental data does not contain values.");
    return messages;
}

SimulationOptionsItem* SimulationView::optionsItem() const
{
    return m_document->simulationOptionsItem();
}

const SampleItem* SimulationView::selectedSampleItem() const
{
    return m_document->sampleModel()->sampleItems().value(m_sampleCombo->currentIndex(), nullptr);
}

const InstrumentItem* SimulationView::selectedInstrumentItem() const
{
    return m_document->instrumentModel()->instrumentItems().value(m_instrumentCombo->currentIndex(),
                                                                  nullptr);
}

const DatafileItem* SimulationView::selectedDatafileItem() const
{
    return m_document->realModel()->realItems().value(m_realDataCombo->currentIndex() - 1,
                                                      nullptr); // -1: "None"
}
