#include "Device/Data/Datafield.h"
#include "Device/IO/IOFactory.h"
#include "GUI/Model/Data/DataItem.h"
#include "GUI/Model/Device/DatafileItem.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Job/JobItem.h"
#include "GUI/Model/Model/JobModel.h"
#include "Tests/GTestWrapper/google_test.h"
#include "Tests/Unit/GUI/Utils.h"
#include <QTest>

TEST(TestJobModel, nonXMLData)
{
    JobModel jobModel;

    // initial state
    EXPECT_EQ(jobModel.dataItems().size(), 0);

    // adding JobItem
    auto* jobItem = jobModel.createJobItem();
    auto* instrumentItem = new GISASInstrumentItem();
    jobItem->copyInstrumentIntoJob(instrumentItem);
    jobItem->createSimulatedDataItem();
    EXPECT_EQ(jobModel.dataItems().size(), 1);

    // adding empty DatafileItem to JobItem
    auto* realItem = jobItem->createDatafileItem();
    EXPECT_EQ(jobModel.dataItems().size(), 1);

    // create specular data in DatafileItem
    realItem->setDatafield(UTest::GUI::createData(0.0, 1));
    EXPECT_EQ(jobModel.dataItems().size(), 2);

    // checking data items of JobModel
    EXPECT_EQ(jobModel.dataItems().indexOf(jobItem->simulatedDataItem()), 0);
    EXPECT_EQ(jobModel.dataItems().indexOf(realItem->dataItem()), 1);

    // replacing the data inside DatafileItem with the data of the same dimensions
    realItem->setDatafield(UTest::GUI::createData(2.0, 1));
    EXPECT_EQ(jobModel.dataItems().size(), 2);

    // replacing the data inside DatafileItem with the data of different dimensions
    const Datafield data = UTest::GUI::createData(3.0, 2);
    EXPECT_THROW(dynamic_cast<DatafileItem*>(realItem)->setDatafield(data), std::runtime_error);

    EXPECT_EQ(jobModel.dataItems().size(), 2);
}

TEST(TestJobModel, saveNonXMLData)
{
    const QString projectDir("DatafieldIOService");
    UTest::GUI::create_dir(projectDir);

    const double value1(1.0), value2(2.0), value3(3.0);

    JobModel jobModel;

    // adding JobItem with instrument
    auto* jobItem = jobModel.createJobItem();
    auto* instrumentItem = new GISASInstrumentItem();
    jobItem->copyInstrumentIntoJob(instrumentItem);

    // create simulated data
    jobItem->createSimulatedDataItem();
    jobItem->simulatedDataItem()->setDatafield(UTest::GUI::createData(value1, 2));

    // create real data
    jobItem->createDatafileItem();
    jobItem->realItem()->setDatafield(UTest::GUI::createData(value2, 2));

    // update names
    jobItem->setJobName("name");
    jobItem->updateFileName();

    // save first time
    jobModel.writeDatafiles(projectDir);
    QTest::qSleep(10);

    // check existence of data on disk
    QString fname1 = "./" + projectDir + "/jobdata_name.int";
    QString fname2 = "./" + projectDir + "/refdata_name.int";
    EXPECT_TRUE(QFile::exists(fname1));
    EXPECT_TRUE(QFile::exists(fname2));

    // read data from disk, check it is the same
    std::unique_ptr<Datafield> dataOnDisk1(
        IO::readData2D(fname1.toStdString(), IO::Filetype2D::bornagain2D));
    std::unique_ptr<Datafield> dataOnDisk2(
        IO::readData2D(fname2.toStdString(), IO::Filetype2D::bornagain2D));
    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk1, *jobItem->simulatedDataItem()->c_field()));
    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk2, *jobItem->realItem()->dataItem()->c_field()));

    // modify data and save the project.
    jobItem->simulatedDataItem()->setDatafield(UTest::GUI::createData(value3, 2));
    jobModel.writeDatafiles(projectDir);
    QTest::qSleep(10);

    // ensure that the simulated data has been changed
    EXPECT_FALSE(UTest::GUI::isTheSame(*dataOnDisk1, *jobItem->simulatedDataItem()->c_field()));
    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk2, *jobItem->realItem()->dataItem()->c_field()));

    // check that data on disk has changed
    dataOnDisk1.reset(IO::readData2D(fname1.toStdString(), IO::Filetype2D::bornagain2D));
    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk1, *jobItem->simulatedDataItem()->c_field()));

    // rename job and check that file on disk changed the name
    jobItem->setJobName("new_job");
    jobItem->updateFileName();
    jobModel.writeDatafiles(projectDir);
    QTest::qSleep(10);

    // check existence of new files on disk
    QString fname1new = "./" + projectDir + "/jobdata_new_job.int";
    QString fname2new = "./" + projectDir + "/refdata_new_job.int";
    EXPECT_TRUE(QFile::exists(fname1new));
    EXPECT_TRUE(QFile::exists(fname2new));

    // check that file with old name was removed
    EXPECT_FALSE(QFile::exists(fname1));
    EXPECT_FALSE(QFile::exists(fname2));
}
